from datetime import datetime, timedelta
import copy

from src.timed_item import TimedItem


time_format = "%d %m %y %H %M"
SECONDS_IN_DAY = 60*60*24
SECONDS_IN_HOUR = 60*60

def list_to_timed_item(my_first_list):
    my_list = copy.deepcopy(my_first_list)
    my_list = read_all_times(my_list)

    return TimedItem(my_list)
    
def read_all_times(my_list):
    my_list[1] = read_delta_maybe_spaces(my_list[1])
    my_list[2] = read_delta_maybe_spaces(my_list[2])
    if my_list[3] != '':
        my_list[3] = read_time(my_list[3])
    if my_list[4] != '':
        my_list[4] = read_time(my_list[4])
    return my_list
    
def read_time(input_string):
    return datetime.strptime(input_string, time_format)
    
def write_time(input_time):
    return datetime.strftime(input_time, time_format)
    
def read_delta_maybe_spaces(input_string):
    input_string = input_string.strip()
    output = read_delta(input_string)
    return output
    
def read_delta(input_string):
    split_by_spaces = input_string.split(" ")
    if only_hours_minutes(split_by_spaces):
        output = read_delta_hours_minutes(input_string)
    elif only_days(split_by_spaces):
        output = timedelta(days=int(split_by_spaces[0]))
    else:
        output = timedelta(days=int(split_by_spaces[0]))
        output += read_delta_hours_minutes(split_by_spaces[1])
    return output

def only_hours_minutes(string_list):
    return len(string_list) == 1 and ":" in string_list[0]
    
def only_days(string_list):
    return len(string_list) == 1 and ":" not in string_list[0]

   
def read_delta_hours_minutes(input_string):
    split_by_colon = input_string.split(":")
    return timedelta(hours=int(split_by_colon[0]), minutes=int(split_by_colon[1]))
    
def write_delta(my_delta):
    all_seconds = my_delta.total_seconds()
    output, all_seconds = get_days(all_seconds)
    output, all_seconds = get_hours(output, all_seconds)
    output, all_seconds = get_minutes(output, all_seconds)
    return output
    
def get_days(all_seconds):
    if all_seconds > SECONDS_IN_DAY:
        output = str(int(all_seconds/SECONDS_IN_DAY))
        all_seconds = all_seconds % SECONDS_IN_DAY
    else:
        output = ''
    return output, all_seconds

def get_hours(output, all_seconds):
    if all_seconds > SECONDS_IN_HOUR:
        output += str(int(all_seconds/SECONDS_IN_HOUR)) + ":"
        all_seconds = all_seconds % SECONDS_IN_HOUR
    else:
        output += "00:"
    return output, all_seconds

def get_minutes(output, all_seconds):
    if all_seconds > 60:
        output += str(int(all_seconds/60))
    else:
        output += "00"
    return output, all_seconds
        


import copy

class Tree():

    def __init__(self, my_dict):
        self.nodes = []
        for item in my_dict.keys():
            my_node = Node(item, my_dict[item])
            self.nodes.append(my_node)
            
    def __eq__(self, other):
        return self.nodes == other.nodes
        
    def __str__(self):
        my_string = "Tree with nodes: "
        for node in self.nodes:
            my_string += str(node) + "\n"
        return my_string
        
    def create_trivial_output(nodes):
        item = LabelledItem((nodes[0].item_list[0][0], "o"))
        output = [item]
        return output
    
    def create_nontrivial_output(nodes):
        output = []
        for item in nodes[0].item_list:
            output.append(LabelledItem(item))
        return output

    def create_output(nodes):
        if len(nodes[0].item_list) == 1:
            output = Tree.create_trivial_output(nodes)
        else:
            output = Tree.create_nontrivial_output(nodes)
        return output

    def do_huffman(self):
        output_string = """"""
        
        while len(self.nodes) > 1:
            self.combine_nodes()
        
        output = []
        if len(self.nodes) == 1:
            output = Tree.create_output(self.nodes)
        return output
        
    def remove_two_lowest_nodes(self, lowest_node, next_lowest_node):
        self.nodes.remove(lowest_node)
        self.nodes.remove(next_lowest_node)
        
    def combine_nodes(self):
        lowest_node, next_lowest_node = self.find_two_lowest_nodes()
        self.remove_two_lowest_nodes(lowest_node, next_lowest_node)
        combined_node = Tree.squish_nodes_together(lowest_node, next_lowest_node)
        self.nodes.append(combined_node)
        
    def find_lowest_node(self):
        min_weight = min([node.weight for node in self.nodes])
        for node in self.nodes:
            if node.weight == min_weight:
                lowest_node = node
        return lowest_node
        
    def find_next_lowest_node(self, lowest_node):
        next_min_weight = min([node.weight for node in self.nodes if node != lowest_node])
        for node in self.nodes:
            if node.weight == next_min_weight and node != lowest_node:
                next_lowest_node = node
        return next_lowest_node
        
    def find_two_lowest_nodes(self):
        lowest_node = self.find_lowest_node()
        next_lowest_node = self.find_next_lowest_node(lowest_node)
        return lowest_node, next_lowest_node   
        
    def squish_nodes_together(lowest_node, next_lowest_node):
        new_node = copy.deepcopy(lowest_node)
        for item in new_node.item_list:
            item[1] = "o" + item[1]
        for item in next_lowest_node.item_list:
            item[1] = "l" + item[1]
        new_node.item_list += next_lowest_node.item_list
        new_node.weight += next_lowest_node.weight
        return new_node

class Node():
    def __init__(self, item, weight):
        self.item_list = [[item, ""]]
        self.weight = weight
        
    def __eq__(self, other):
        return self.item_list == other.item_list and self.weight == other.weight
        
    def __str__(self):
        return f"Node with items {self.item_list} and weight {self.weight}"
      
      
class LabelledItem():

    def __init__(self, item):
         self.label = item[1]
         self.name = item[0]
         
    def __str__(self):
        return f"Item with label {self.label} and name {self.name}"
        
    def __eq__(self, other):
        return self.label == other.label and self.name == other.name
  


import sys
import datetime

from src.scheduler import Scheduler
from src.timed_schedule_config import TimedScheduleConfig
from src.file_persistence import FilePersistence

config_file = sys.argv[1]
persistence_file = sys.argv[2]
        
my_persistence = FilePersistence(config_file, persistence_file)

my_input = ""
while my_input not in [str(n) for n in range(1, 6)]:
    my_input = input("How overwhelmed are you from 1-5?\n")
        
timed_config = TimedScheduleConfig(my_persistence, int(my_input))

my_schedule = Scheduler(timed_config)

my_item, do_feedback = my_schedule.pick_item(datetime.datetime.now())

print(my_item)

if do_feedback:
    timed_config.get_feedback(my_item)







class TimeIntervalNotSpecifiedException(Exception):
    pass
    
class BadChangeTypeException(Exception):
    pass
    
class NoSuchItemNameException(Exception):
    pass
    
class HangsOnSearchingLoopException(Exception):
    pass


import csv
import datetime

from src.schedule_exceptions import BadChangeTypeException, TimeIntervalNotSpecifiedException

MAX_WEIGHT = 1000
HIGHEST_OVERWHELM = 5

DONE = "DONE"
SNOOZE = "SNOOZE"

class TimedScheduleConfig():

    def __init__(self, persistence, overwhelm=0):
        self.persistence = persistence
        TimedScheduleConfig.fix_lines(persistence)
        self.overwhelm = overwhelm
        
    def create_weighted_items(self, input_time=datetime.datetime.now()):
        weights = {}
        for item in self.list_timed_items():
            weight = self.get_weight(item, input_time)
            if weight != 0:
                weights[item.name] = self.get_weight(item, input_time)
        self.weights = weights
        
    def list_timed_items(self):
        return self.persistence.list_timed_items()
        
    def fix_lines(persistence):
        for idx, my_list in enumerate(persistence.all_lists):
            while len(my_list) < 8:
                my_list.append("")
            persistence.all_lists[idx] = my_list
            if my_list[1] == "" or my_list[2] == "":
                raise TimeIntervalNotSpecifiedException
     
    def get_weight(self, item, input_time):
        if not item.done:
            output =  MAX_WEIGHT
        else:
            output = self.get_weight_maybe_due(item, input_time)
        if item.snooze:
            if input_time < item.snooze:
                output = 0
        if not type(item.overwhelm) == str:
            if HIGHEST_OVERWHELM - self.overwhelm + 1 < item.overwhelm:
                output = 0
        return output
        
    def get_weight_maybe_due(self, item, input_time):
        due_time = item.done + item.wait_interval_start
        if input_time < due_time:
            output = 0
        else:
            slope = self.get_slope(item)
            output = slope * type(self.persistence).number((input_time - due_time))
        return output
        
    def get_slope(self, item):
        return MAX_WEIGHT/type(self.persistence).number((item.wait_interval_end - item.wait_interval_start))
        
    def find_in_list_timed_items(self, item_name):
        for idx, item in enumerate(self.list_timed_items()):
            if item.name == item_name:
                output_item = item
                output_idx = idx
        return output_idx, output_item
                
    def done(self, item_name, input_time):
        idx, item = self.find_in_list_timed_items(item_name)
        TimedScheduleConfig.set_done(idx, item, self.persistence, input_time)
        
    def set_done(idx, item, persistence, input_time):
        persistence.all_lists[idx][3] = type(persistence).time_to_string(input_time)
        persistence.persist_done(idx, input_time)
        
    def snooze(self, item_name, input_time):
        idx, item = self.find_in_list_timed_items(item_name)
        TimedScheduleConfig.set_snooze(idx, item, self.persistence, input_time)
        
    def set_snooze(idx, item, persistence, input_time):
        if item.snooze_amount and TimedScheduleConfig.not_is_string_and_space(item.snooze_amount) and item.snooze_amount != "":
            persistence.snooze_amount(idx, input_time)
        else:
            persistence.snooze_default(idx, input_time)
        persistence.persist_snooze(idx, input_time)
        
    def set_item_overwhelm(self, item_name, item_overwhelm):
        idx, item = self.find_in_list_timed_items(item_name)
        TimedScheduleConfig.set_overwhelm(idx, item, item_overwhelm, self.persistence)
        
    def set_overwhelm(idx, item, item_overwhelm, persistence):
        persistence.all_lists[idx][6] = item_overwhelm
        persistence.persist_overwhelm(idx, item_overwhelm)
    
    def not_is_string_and_space(stringie):
        output = True
        if type(stringie) == str and stringie.isspace():
            output = False
        return output
        
    def get_feedback(self, item_name, change_type=None, input_time=None, item_overwhelm=None):
        if not change_type:
            change_type = type(self.persistence).get_user_input()
        if not input_time:
            input_time = type(self.persistence).get_input_time()
        if not item_overwhelm:
            item_overwhelm = type(self.persistence).get_item_overwhelm()
        self.set_item_overwhelm(item_name, item_overwhelm)
        if change_type == DONE:
            self.done(item_name, input_time)
        elif change_type == SNOOZE:
            self.snooze(item_name, input_time)
        else:
            raise BadChangeTypeException
        


from src.schedule_exceptions import TimeIntervalNotSpecifiedException

class TimedItem():
    def __init__(self, list_item):
        self.name = list_item[0]
        self.wait_interval_start = list_item[1]
        self.wait_interval_end = list_item[2]
        self.done = list_item[3]
        self.snooze = list_item[4]
        if list_item[6] != "":
            self.overwhelm = int(list_item[6])
        else:
            self.overwhelm = list_item[6]
        self.snooze_amount = list_item[5]
        
       
    def __eq__(self, other):
        return str(self) == str(other)
        
    def __str__(self):
        return f"{self.name}, {self.wait_interval_start}, {self.wait_interval_end}, {self.done}, {self.snooze}"
            
    def set_done(self, input_time):
        self.done = input_time
        
    def set_snooze(self, input_time, snooze_amount):
        self.snooze = input_time + snooze_amount



import csv
import copy
import sys
import random
import math
import timeit

from src.schedule_exceptions import HangsOnSearchingLoopException 
from src.tree import Tree

class Huffman():
    def __init__(self, my_input):
        self.input_tree = Tree(my_input)
        self.output_tree = copy.deepcopy(self.input_tree)
        
    def do_huffman(self):
        output = self.output_tree.do_huffman()
        self.labelled_items = output
        
    def pick_item(self):
        random_number = random.randint(0, sys.maxsize)
        return self.get_tree_element(random_number)
        
    def initialize(the_number):
        if the_number == 0:
            output_string = "o"
        else:
            output_string = ""
        return output_string
        
    def lowest_binary_digit(the_number):
        if the_number % 2 == 1:
            output = "l"
        else:
            output = "o"
        return output
        
    def remove_lowest_binary_digit(the_number):
        return math.floor(the_number/2)
        
    def convert_next_binary_digit(output_string, remaining_to_be_converted):
        output_string = Huffman.lowest_binary_digit(remaining_to_be_converted) + output_string
        remaining_to_be_converted = Huffman.remove_lowest_binary_digit(remaining_to_be_converted)
        return output_string, remaining_to_be_converted

    def get_desired(the_number):
        remaining_to_be_converted = the_number
        output_string = Huffman.initialize(remaining_to_be_converted)
        while remaining_to_be_converted >= 1:
            output_string, remaining_to_be_converted = Huffman.convert_next_binary_digit(output_string, remaining_to_be_converted)
        return output_string
        
    def compare_low_bits(desired_end, encoding_begin, encoding_end):
        return (encoding_end == desired_end) and ("l" not in encoding_begin)
        
    def compare_strings(encoding, desired):
            n = min(len(encoding), len(desired))
            desired_end = desired[-n:]
            encoding_begin = encoding[:-n]
            encoding_end = encoding[-n:]
            matched = Huffman.compare_low_bits(desired_end, encoding_begin, encoding_end)
            return matched       

    def check_item(item, desired):
        encoding = item.label
        haveFoundNext = Huffman.compare_strings(encoding, desired)
        return haveFoundNext
        
    def do_searching_loop(labelled_items, desired):
        idx = 0
        output = None
        while idx < len(labelled_items) and not output:
            item = labelled_items[idx]
            haveFoundNext = Huffman.check_item(item, desired)
            if haveFoundNext:
                output = item.name
            idx = idx + 1
        return output

    def find_in_labelled_items(labelled_items, my_number):
        desired = Huffman.get_desired(my_number)
        output = Huffman.do_searching_loop(labelled_items, desired)
        return output 

    def run_loop(labelled_items, my_number, my_time_tracker):
        total_time = my_time_tracker.total_time()
        output = None
        while not output and total_time < my_time_tracker.time_limit:
            output = Huffman.find_in_labelled_items(labelled_items, my_number)
            my_number = my_number + 1 
            total_time = my_time_tracker.total_time()
        return output, total_time
                   
    def get_tree_element(self, my_number, my_time_tracker=None):
        if not my_time_tracker:
            my_time_tracker = TimeTracker()
        output, total_time = Huffman.run_loop(self.labelled_items, my_number, my_time_tracker)
        if total_time >= my_time_tracker.time_limit:
            raise HangsOnSearchingLoopException
        return output

        
class TimeTracker():

    def default_timer():
        while True:
            yield timeit.default_timer()

    def __init__(self, timer=None, time_limit=.5):
        if not timer:
            self.timer = TimeTracker.default_timer
        else:
            self.timer = timer
        self.time_limit = time_limit
        self.time_generator = self.timer()
        self.initial_time = next(self.time_generator)
        
    def total_time(self):
        next_guy = next(self.time_generator)
        total_time = next_guy - self.initial_time
        return total_time
    


from src.timed_schedule_config import TimedScheduleConfig
from src.huffman_tree import Huffman

class Scheduler:
    
    def __init__(self, timed_config):
        self.config = timed_config
                
    
    def pick_item(self, input_time=0):
        self.config.create_weighted_items(input_time)
        all_the_items = list(self.config.weights.keys())
        if len(all_the_items) == 0:
            return "Nothing to do!  Wait a few minutes before trying again.", False
        if len(all_the_items) == 1:
            return all_the_items[0], True
        else:
            return self.pick_from_longer_list(input_time)
            
            
    def pick_from_longer_list(self, input_time):
        self.create_weighted_items(input_time)
        self.do_huffman()
        return self.schedule.pick_item(), True
        
    def create_weighted_items(self, input_time):
        self.config.create_weighted_items(input_time)
        
    def do_huffman(self):
        self.schedule = Huffman(self.config.weights)
        self.schedule.do_huffman()




import datetime

from src.schedule_parser import list_to_timed_item, time_format, read_delta_maybe_spaces 
from src.schedule_exceptions import BadChangeTypeException, NoSuchItemNameException


class FilePersistence:
    def __init__(self, input_filename, output_filename):
        self.output_filename = output_filename
        with open(input_filename, 'r') as f:
            all_lines = f.readlines()
        self.all_lists = FilePersistence.fill_in_all_lists(all_lines)
        
    def fill_in_all_lists(all_lines):
        output = []
        for line in all_lines:
            my_list = line.split(",")[:-1]
            if not FilePersistence.is_trivial(my_list):
                output.append(my_list)
        return output
    
    def is_trivial(my_list):
        return my_list[0] == ""
        
    def list_timed_items(self):
        output = []
        for my_list in self.all_lists:
            timed_item = list_to_timed_item(my_list)
            output.append(timed_item)
        return output
        
    def number(date_input):
        return date_input.total_seconds()/60
            
    def snooze_default(self, idx, input_time):
        my_item = self.all_lists[idx]
        my_item[4] = datetime.datetime.strftime(input_time + datetime.timedelta(hours=1), time_format)
        self.all_lists[idx] = my_item
        
    def snooze_amount(self, idx, input_time):
        my_item = self.all_lists[idx]
        my_item[4] = datetime.datetime.strftime(input_time + read_delta_maybe_spaces(my_item[5]), time_format)
        
    def persist_done(self, splitsies, input_time):
        self.write_all_lists()
        
    def persist_snooze(self, splitsies, input_time):
        self.write_all_lists()
        
    def persist_overwhelm(self, splitsies, item_overwhelm):
        self.write_all_lists()
        
    def time_to_string(date):
        return datetime.datetime.strftime(date, time_format)
        
    def write_all_lists(self):
        with open(self.output_filename, 'w') as f:
            for my_list in self.all_lists:
                f.write(",".join(my_list))
                f.write("\n")
                
    def get_input_time():
        return datetime.datetime.now()
        
    def get_user_input(my_input=None):
        while my_input not in ["1", "2"]:
            my_input = input("Enter 1 to mark as done, enter 2 to snooze\n")
        return FilePersistence.parse_user_input(my_input)
        
    def parse_user_input(my_input):
        my_input_dict = {'1': "DONE", '2': "SNOOZE"}
        return my_input_dict[my_input]
        
    def get_item_overwhelm(my_input=None):
        while my_input not in [str(n) for n in list(range(1, 6))]:
            my_input = input("Rate how overwhelming this item is from 1-5\n")
        return my_input